@@ -19,17 +19,17 @@
 #' @param label (optional) Label of the future (where applicable, becomes the
 #' job name for most job schedulers).
 #'
-#' @param resources (optional) A named list passed to the batchtools template
-#' (available as variable `resources`).
+#' @param resources (optional) A named list passed to the \pkg{batchtools}
+#' template (available as variable `resources`).  See Section 'Resources'
+#' in [batchtools::submitJobs()] more details.
 #'
 #' @param workers (optional) The maximum number of workers the batchtools
 #' backend may use at any time.   Interactive and "local" backends can only
 #' process one future at the time (`workers = 1L`), whereas HPC backends,
 #' where futures are resolved via separate jobs on a scheduler, can have
 #' multiple workers.  In the latter, the default is `workers = NULL`, which
-#' will resolve to `getOption("future.batchtools.workers")`.  If that is not
-#' specified, the value of environment variable `R_FUTURE_BATCHTOOLS_WORKERS`
-#' will be used.  If neither are specified, then the default is `100`.
+#' will resolve to `getOption("future.batchtools.workers")`.  If neither
+#' are specified, then the default is `100`.
 #'
 #' @param finalize If TRUE, any underlying registries are
 #' deleted when this object is garbage collected, otherwise not.
@@ -48,7 +48,6 @@
 #'
 #' @export
 #' @importFrom future Future getGlobalsAndPackages
-#' @importFrom batchtools submitJobs
 #' @keywords internal
 BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
                              substitute = TRUE,
@@ -62,8 +61,59 @@ BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
                              registry = list(),
                              ...) {
   if (substitute) expr <- substitute(expr)
+  assert_no_positional_args_but_first()
 
-  if (!is.null(label)) label <- as.character(label)
+  ## Record globals
+  gp <- getGlobalsAndPackages(expr, envir = envir, globals = globals)
+
+  future <- Future(expr = gp$expr, envir = envir, substitute = FALSE,
+                   globals = gp$globals,
+                   packages = unique(c(packages, gp$packages)),
+                   label = label,
+                   ...)
+
+  future <- as_BatchtoolsFuture(future,
+                                resources = resources,
+                                workers = workers,
+                                finalize = finalize,
+                                conf.file = conf.file,
+                                cluster.functions = cluster.functions,
+                                registry = registry)
+
+  future
+}
+
+
+## Helper function to create a BatchtoolsFuture from a vanilla Future
+#' @importFrom utils file_test
+as_BatchtoolsFuture <- function(future,
+                                resources = list(),
+                                workers = NULL,
+                                finalize = getOption("future.finalize", TRUE),
+                                conf.file = findConfFile(),
+                                cluster.functions = NULL,
+                                registry = list(),
+                                ...) {
+  if (is.function(workers)) workers <- workers()
+  if (is.null(workers)) {
+    workers <- getOption("future.batchtools.workers", default = 100)
+    stop_if_not(
+      is.numeric(workers),
+      length(workers) == 1,
+      !is.na(workers), workers >= 1
+    )
+  } else {
+    stop_if_not(length(workers) >= 1)
+    if (is.numeric(workers)) {
+      stop_if_not(length(workers) == 1, !is.na(workers), workers >= 1)
+    } else if (is.character(workers)) {
+      stop_if_not(length(workers) >= 0, !anyNA(workers))
+    } else {
+      stop("Argument 'workers' should be either a numeric or a function: ",
+           mode(workers))
+    }
+  }
+  future$workers <- workers
 
   if (!is.null(cluster.functions)) {
     stop_if_not(is.list(cluster.functions))
@@ -79,17 +129,6 @@ BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
     }
   }
   
-  if (is.function(workers)) workers <- workers()
-  if (!is.null(workers)) {
-    stop_if_not(length(workers) >= 1)
-    if (is.numeric(workers)) {
-      stop_if_not(!anyNA(workers), all(workers >= 1))
-    } else {
-      stop("Argument 'workers' should be either a numeric or a function: ",
-           mode(workers))
-    }
-  }
-
   stop_if_not(is.list(registry))
   if (length(registry) > 0L) {
     stopifnot(!is.null(names(registry)), all(nzchar(names(registry))))
@@ -97,39 +136,18 @@ BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
   
   stop_if_not(is.list(resources))
 
-  ## Record globals
-  gp <- getGlobalsAndPackages(expr, envir = envir, globals = globals)
-
-  future <- Future(expr = gp$expr, envir = envir, substitute = FALSE,
-                   workers = workers, label = label,
-                   version = "1.8", .callResult = TRUE,
-                   ...)
-
-  future$globals <- gp$globals
-  future$packages <- unique(c(packages, gp$packages))
-
-  ## Create batchtools registry
-  reg <- temp_registry(
-    label = future$label,
+  ## batchtools configuration
+  future$config <- list(
+    reg = NULL,
+    jobid = NA_integer_,
+    resources = resources,
     conf.file = conf.file,
     cluster.functions = cluster.functions,
-    config = registry
+    registry = registry,
+    finalize = finalize
   )
-  debug <- getOption("future.debug", FALSE)
-  if (debug) mprint(reg)
 
-  ## batchtools configuration
-  config <- list(reg = reg, jobid = NA_integer_,
-                 resources = resources)
-
-  future$config <- config
-
-  future <- structure(future, class = c("BatchtoolsFuture", class(future)))
-
-  ## Register finalizer?
-  if (finalize) future <- add_finalizer(future)
-
-  future
+  structure(future, class = c("BatchtoolsFuture", class(future)))
 }
 
 
@@ -140,15 +158,25 @@ BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
 #'
 #' @export
 #' @keywords internal
-print.BatchtoolsFuture <- function(x, ...) {
+print.BatchtoolsFuture <- function(x, ...) {  
   NextMethod()
 
   ## batchtools specific
-  reg <- x$config$reg
+  config <- x$config
 
-  ## Type of batchtools future
-  printf("batchtools cluster functions: %s\n",
-         sQuote(reg$cluster.functions$name))
+  conf.file <- config$conf.file
+  printf("batchtools configuration file: %s\n", file_info(conf.file))
+  
+  reg <- config$reg
+  if (inherits(reg, "Registry")) {
+    cluster.functions <- reg$cluster.functions
+    printf("batchtools cluster functions: %s\n",
+           sQuote(cluster.functions$name))
+    template <- attr(cluster.functions, "template")
+    printf("batchtools cluster functions template: %s\n", file_info(template))
+  } else {
+    printf("batchtools cluster functions: <none>\n")
+  }
 
   ## Ask for status once
   status <- status(x)
@@ -161,39 +189,28 @@ print.BatchtoolsFuture <- function(x, ...) {
     printf("batchtools %s: Not found (happens when finished and deleted)\n",
            class(reg))
   } else {
-    printf("batchtools Registry:\n")
-    printf("  File dir exists: %s\n", file_test("-d", reg$file.dir))
-    printf("  Work dir exists: %s\n", file_test("-d", reg$work.dir))
-    try(print(reg))
+    if (inherits(reg, "Registry")) {
+      printf("batchtools Registry:\n")
+      printf("  File dir exists: %s\n", file_test("-d", reg$file.dir))
+      printf("  Work dir exists: %s\n", file_test("-d", reg$work.dir))
+      try(print(reg))
+    } else {
+      printf("batchtools Registry: <NA>\n")
+    }
   }
 
   invisible(x)
 }
 
 
-status <- function(...) UseMethod("status")
-finished <- function(...) UseMethod("finished")
-loggedError <- function(...) UseMethod("loggedError")
-loggedOutput <- function(...) UseMethod("loggedOutput")
-
-#' Status of batchtools future
-#'
-#' @param future The future.
-#' @param \ldots Not used.
-#'
-#' @return A character vector or a logical scalar.
-#'
-#' @aliases status finished result
-#'          loggedError loggedOutput
-#' @keywords internal
-#'
-#' @export
-#' @export status
-#' @export finished
-#' @export loggedError
-#' @export loggedOutput
 #' @importFrom batchtools getStatus
-status.BatchtoolsFuture <- function(future, ...) {
+status <- function(future, ...) {
+  debug <- getOption("future.debug", FALSE)
+  if (debug) {
+    mdebug("status() for ", class(future)[1], " ...")
+    on.exit(mdebug("status() for ", class(future)[1], " ... done"), add = TRUE)
+  }
+  
   ## WORKAROUND: Avoid warnings on partially matched arguments
   get_status <- function(...) {
     ## Temporarily disable batchtools output?
@@ -206,7 +223,11 @@ status.BatchtoolsFuture <- function(future, ...) {
       oopts <- list()
     }
     on.exit(options(oopts))
-    batchtools::getStatus(...)
+    ## WORKAROUND: batchtools::getStatus() updates the RNG state,
+    ## which we must make sure to undo.
+    with_stealth_rng({
+      batchtools::getStatus(...)
+    })
   } ## get_status()
 
   config <- future$config
@@ -230,20 +251,38 @@ status.BatchtoolsFuture <- function(future, ...) {
     if (result_has_errors(result)) status <- unique(c("error", status))
   }
 
+  if (debug) mdebug("- status: ", paste(sQuote(status), collapse = ", "))
+
   status
 }
 
 
-#' @export
-#' @keywords internal
-finished.BatchtoolsFuture <- function(future, ...) {
+finished <- function(future, ...) {
   status <- status(future)
   if (is_na(status)) return(NA)
   any(c("finished", "error", "expired") %in% status)
 }
 
-#' @export
+
+
+#' Logged output of batchtools future
+#'
+#' @param future The future.
+#' @param \ldots Not used.
+#'
+#' @return A character vector or a logical scalar.
+#'
+#' @aliases loggedOutput loggedError
+#'
+#' @export loggedError
+#' @export loggedOutput
 #' @keywords internal
+loggedOutput <- function(...) UseMethod("loggedOutput")
+loggedError <- function(...) UseMethod("loggedError")
+
+
+#' @importFrom batchtools getErrorMessages
+#' @export
 loggedError.BatchtoolsFuture <- function(future, ...) {
   stat <- status(future)
   if (is_na(stat)) return(NULL)
@@ -259,6 +298,7 @@ loggedError.BatchtoolsFuture <- function(future, ...) {
 
   config <- future$config
   reg <- config$reg
+  if (!inherits(reg, "Registry")) return(NULL)
   jobid <- config$jobid
   res <- getErrorMessages(reg = reg, ids = jobid)  ### CHECKED
   msg <- res$message
@@ -269,7 +309,6 @@ loggedError.BatchtoolsFuture <- function(future, ...) {
 
 #' @importFrom batchtools getLog
 #' @export
-#' @keywords internal
 loggedOutput.BatchtoolsFuture <- function(future, ...) {
   stat <- status(future)
   if (is_na(stat)) return(NULL)
@@ -283,6 +322,7 @@ loggedOutput.BatchtoolsFuture <- function(future, ...) {
 
   config <- future$config
   reg <- config$reg
+  if (!inherits(reg, "Registry")) return(NULL)
   jobid <- config$jobid
   getLog(id = jobid, reg = reg)
 } # loggedOutput()
@@ -327,7 +367,7 @@ result.BatchtoolsFuture <- function(future, cleanup = TRUE, ...) {
   if (is_na(stat)) {
     label <- future$label
     if (is.null(label)) label <- "<none>"
-    stop(sprintf("The result no longer exists (or never existed) for Future ('%s') of class %s", label, paste(sQuote(class(future)), collapse = ", "))) #nolint
+    stopf("The result no longer exists (or never existed) for Future ('%s') of class %s", label, paste(sQuote(class(future)), collapse = ", ")) #nolint
   }
 
   result <- await(future, cleanup = FALSE)
@@ -342,13 +382,15 @@ result.BatchtoolsFuture <- function(future, cleanup = TRUE, ...) {
 
 
 #' @importFrom future run getExpression
-#' @importFrom batchtools batchExport batchMap saveRegistry setJobNames
+#' @importFrom batchtools batchExport batchMap saveRegistry setJobNames submitJobs
+#' @importFrom utils capture.output str
 #' @export
 run.BatchtoolsFuture <- function(future, ...) {
   if (future$state != "created") {
     label <- future$label
     if (is.null(label)) label <- "<none>"
-    stop(sprintf("A future ('%s') can only be launched once.", label))
+    msg <- sprintf("A future ('%s') can only be launched once.", label)
+    stop(FutureError(msg, future = future))
   }
 
   ## Assert that the process that created the future is
@@ -372,7 +414,29 @@ run.BatchtoolsFuture <- function(future, ...) {
   ## Always evaluate in local environment
   expr <- substitute(local(expr), list(expr = expr))
 
+  ## (i) Create batchtools registry
   reg <- future$config$reg
+  stop_if_not(is.null(reg) || inherits(reg, "Registry"))
+  if (is.null(reg)) {
+    if (debug) mprint("- Creating batchtools registry")
+    config <- future$config
+    stop_if_not(is.list(config))
+    
+    ## Create batchtools registry
+    reg <- temp_registry(
+      label             = future$label,
+      conf.file         = config$conf.file,
+      cluster.functions = config$cluster.functions,
+      config            = config$registry
+    )
+    if (debug) mprint(reg)
+    future$config$reg <- reg
+
+    ## Register finalizer?
+    if (config$finalize) future <- add_finalizer(future)
+    
+    config <- NULL
+  }
   stop_if_not(inherits(reg, "Registry"))
 
   ## (ii) Attach packages that needs to be attached
@@ -405,16 +469,20 @@ run.BatchtoolsFuture <- function(future, ...) {
 
   ## 1. Add to batchtools for evaluation
   mdebug("batchtools::batchMap()")
-  jobid <- batchMap(fun = geval, list(expr),
-                    more.args = list(substitute = TRUE), reg = reg)
+  ## WORKAROUND: batchtools::batchMap() updates the RNG state,
+  ## which we must make sure to undo.
+  with_stealth_rng({
+    jobid <- batchMap(fun = geval, list(expr),
+                      more.args = list(substitute = TRUE), reg = reg)
+  })
 
-  ## 1b. Set job name, if specified
+  ## 2. Set job name, if specified
   label <- future$label
   if (!is.null(label)) {
     setJobNames(ids = jobid, names = label, reg = reg)
   }
   
-  ## 2. Update
+  ## 3. Update
   future$config$jobid <- jobid
   mdebugf("Created %s future #%d", class(future)[1], jobid$job.id)
 
@@ -440,52 +508,47 @@ run.BatchtoolsFuture <- function(future, ...) {
     }, error = function(ex) list())
   }
 
-  ## 3. Submit
+  ## 4. Wait for an available worker
+  waitForWorker(future, workers = future$workers)
+
+  ## 5. Submit
   future$state <- "running"
   resources <- future$config$resources
   if (is.null(resources)) resources <- list()
 
-  batchtools::submitJobs(reg = reg, ids = jobid, resources = resources)
+  ## WORKAROUND: batchtools::submitJobs() updates the RNG state,
+  ## which we must make sure to undo.
+  tryCatch({
+    with_stealth_rng({
+      submitJobs(reg = reg, ids = jobid, resources = resources)
+    })
+  }, error = function(ex) {
+    msg <- conditionMessage(ex)
+    label <- future$label
+    if (is.null(label)) label <- "<none>"
+    msg <- sprintf("Failed to submit %s (%s). The reason was: %s", class(future)[1], label, msg)
+    info <- capture.output(str(resources))
+    info <- paste(info, collapse = "\n")
+    msg <- sprintf("%s\nTROUBLESHOOTING INFORMATION:\nbatchtools::submitJobs() was called with the following 'resources' argument:\n%s\n", msg, info)
+    stop(BatchtoolsFutureError(msg, future = future))
+  })
 
   mdebugf("Launched future #%d", jobid$job.id)
 
+  ## 6. Rerserve worker for future
+  registerFuture(future)
+
   invisible(future)
 } ## run()
 
 
-await <- function(...) UseMethod("await")
-
-#' Awaits the value of a batchtools future
-#'
-#' @param future The future.
-#' @param cleanup If TRUE, the registry is completely removed upon
-#' success, otherwise not.
-#' @param timeout Total time (in seconds) waiting before generating an error.
-#' @param delta The number of seconds to wait between each poll.
-#' @param alpha A factor to scale up the waiting time in each iteration such
-#' that the waiting time in the k:th iteration is `alpha ^ k * delta`.
-#' @param \ldots Not used.
-#'
-#' @return The value of the evaluated expression.
-#' If an error occurs, an informative Exception is thrown.
-#'
-#' @details
-#' Note that `await()` should only be called once, because
-#' after being called the actual asynchronous future may be removed
-#' and will no longer available in subsequent calls.  If called
-#' again, an error may be thrown.
-#'
-#' @export
-#' @importFrom batchtools getErrorMessages loadResult waitForJobs
+#' @importFrom batchtools loadResult waitForJobs
 #' @importFrom utils tail
-#' @keywords internal
-await.BatchtoolsFuture <- function(future, cleanup = TRUE,
-                                   timeout = getOption("future.wait.timeout",
-                                                       30 * 24 * 60 * 60),
-                                   delta = getOption("future.wait.interval",
-                                                     1.0),
-                                   alpha = getOption("future.wait.alpha", 1.01),
-                                   ...) {
+await <- function(future, cleanup = TRUE,
+                  timeout = getOption("future.wait.timeout", 30 * 24 * 60 * 60),
+                  delta = getOption("future.wait.interval", 1.0),
+                  alpha = getOption("future.wait.alpha", 1.01),
+                  ...) {
   stop_if_not(is.finite(timeout), timeout >= 0)
   stop_if_not(is.finite(alpha), alpha > 0)
   
@@ -494,6 +557,7 @@ await.BatchtoolsFuture <- function(future, cleanup = TRUE,
   expr <- future$expr
   config <- future$config
   reg <- config$reg
+  stop_if_not(inherits(reg, "Registry"))
   jobid <- config$jobid
 
   mdebug("batchtools::waitForJobs() ...")
@@ -523,10 +587,18 @@ await.BatchtoolsFuture <- function(future, cleanup = TRUE,
     label <- future$label
     if (is.null(label)) label <- "<none>"
     if ("finished" %in% stat) {
+      mdebug("- batchtools::loadResult() ...")
       result <- loadResult(reg = reg, id = jobid)
+      mdebug("- batchtools::loadResult() ... done")
       if (inherits(result, "FutureResult")) {
         prototype_fields <- c(prototype_fields, "batchtools_log")
         result[["batchtools_log"]] <- try({
+          mdebug("- batchtools::getLog() ...")
+          on.exit(mdebug("- batchtools::getLog() ... done"))
+	  ## Since we're already collected the results, the log file
+	  ## should already exist, if it exists.  Because of this,
+	  ## only poll for the log file for a second before giving up.
+	  reg$cluster.functions$fs.latency <- 1.0
           getLog(id = jobid, reg = reg)
         }, silent = TRUE)
         if (result_has_errors(result)) cleanup <- FALSE
@@ -556,7 +628,7 @@ await.BatchtoolsFuture <- function(future, cleanup = TRUE,
       msg <- sprintf("BatchtoolsDeleted: Cannot retrieve value. Future ('%s') deleted: %s", label, reg$file.dir) #nolint
       stop(BatchtoolsFutureError(msg, future = future))
     }
-    if (debug) { mstr(result) }
+    if (debug) { mstr(result) }    
   } else {
     cleanup <- FALSE
     msg <- sprintf("AsyncNotReadyError: Polled for results for %s seconds every %g seconds, but asynchronous evaluation for future ('%s') is still running: %s", timeout, delta, label, reg$file.dir) #nolint
@@ -613,6 +685,10 @@ delete.BatchtoolsFuture <- function(future,
   ## Identify registry
   config <- future$config
   reg <- config$reg
+  
+  ## Trying to delete a non-launched batchtools future?
+  if (!inherits(reg, "Registry")) return(invisible(TRUE))
+  
   path <- reg$file.dir
 
   ## Already deleted?
@@ -651,6 +727,9 @@ delete.BatchtoolsFuture <- function(future,
   result <- result(future, cleanup = FALSE)
   stop_if_not(inherits(result, "FutureResult"))
 
+  ## Free up worker
+  unregisterFuture(future)
+
   ## To simplify post mortem troubleshooting in non-interactive sessions,
   ## should the batchtools registry files be removed or not?
   mdebugf("delete(): Option 'future.delete = %s",
@@ -683,15 +762,19 @@ delete.BatchtoolsFuture <- function(future,
   on.exit(options(oopts))
 
   ## Try to delete registry
-  interval <- delta
-  for (kk in seq_len(times)) {
-    try(clearRegistry(reg = reg), silent = TRUE)
-    try(removeRegistry(wait = 0.0, reg = reg), silent = FALSE)
-    if (!file_test("-d", path)) break
-    Sys.sleep(interval)
-    interval <- alpha * interval
-  }
-
+  ## WORKAROUND: batchtools::clearRegistry() and
+  ## batchtools::removeRegistry() update the RNG state,
+  ## which we must make sure to undo.
+  with_stealth_rng({
+    interval <- delta
+    for (kk in seq_len(times)) {
+      try(clearRegistry(reg = reg), silent = TRUE)
+      try(removeRegistry(wait = 0.0, reg = reg), silent = FALSE)
+      if (!file_test("-d", path)) break
+      Sys.sleep(interval)
+      interval <- alpha * interval
+    }
+  })
 
   ## Success?
   if (file_test("-d", path)) {
@@ -715,16 +798,37 @@ delete.BatchtoolsFuture <- function(future,
 
 add_finalizer <- function(...) UseMethod("add_finalizer")
 
-add_finalizer.BatchtoolsFuture <- function(future, ...) {
+add_finalizer.BatchtoolsFuture <- function(future, debug = FALSE, ...) {
   ## Register finalizer (will clean up registries etc.)
 
-  reg.finalizer(future, f = function(gcenv) {
-    if (inherits(future, "BatchtoolsFuture") &&
-        "future.batchtools" %in% loadedNamespaces()) {
-      try({
-        delete(future, onRunning = "skip", onMissing = "ignore",
-               onFailure = "warning")
+  if (debug) {
+    mdebug("add_finalizer() for ", sQuote(class(future)[1]), " ...")
+    on.exit(mdebug("add_finalizer() for ", sQuote(class(future)[1]), " ... done"), add = TRUE)
+  }
+
+  reg.finalizer(future, f = function(f) {
+    if (debug) {
+      if (!exists("mdebug", mode = "function")) mdebug <- message
+      mdebug("Finalize ", sQuote(class(f)[1]), " ...")
+      on.exit(mdebug("Finalize ", sQuote(class(f)[1]), " ... done"), add = TRUE)
+    }
+    if (inherits(f, "BatchtoolsFuture") && "future.batchtools" %in% loadedNamespaces()) {
+      if (debug) {
+        mdebug("- attempting to delete future")
+        if (requireNamespace("utils", quietly = TRUE)) {
+          mdebug(utils::capture.output(utils::str(as.list(f))))
+        }
+      }
+      res <- try({
+        delete(f, onRunning = "skip", onMissing = "ignore", onFailure = "warning")
       })
+      if (debug) {
+        if (inherits(res, "try-error")) {
+          mdebug("- Failed to delete: ", sQuote(res))
+        } else {
+          mdebug("- deleted: ", res)
+        }
+      }
     }
   }, onexit = TRUE)
 
