@@ -18,8 +18,7 @@ stop_if_not <- function(...) {
         mc <- match.call()
         call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
         if (length(call) > 1L) call <- paste(call[1L], "....")
-        stop(sprintf("%s is not TRUE", sQuote(call)),
-             call. = FALSE, domain = NA)
+        stopf("%s is not TRUE", sQuote(call), call. = FALSE, domain = NA)
     }
   }
   
@@ -47,23 +46,35 @@ capture_output <- function(expr, envir = parent.frame(), ...) {
 
 printf <- function(...) cat(sprintf(...))
 
-mcat <- function(...) message(..., appendLF = FALSE)
-
-mprintf <- function(...) message(sprintf(...), appendLF = FALSE)
-
-mprint <- function(...) {
-  bfr <- capture_output(print(...))
-  bfr <- paste(c(bfr, ""), collapse = "\n")
-  message(bfr, appendLF = FALSE)
+now <- function(x = Sys.time(), format = "[%H:%M:%OS3] ") {
+  ## format(x, format = format) ## slower
+  format(as.POSIXlt(x, tz = ""), format = format)
 }
 
-#' @importFrom utils str
-mstr <- function(...) {
-  bfr <- capture_output(str(...))
-  bfr <- paste(c(bfr, ""), collapse = "\n")
-  message(bfr, appendLF = FALSE)
+mdebug <- function(..., debug = getOption("future.debug", FALSE)) {
+  if (!debug) return()
+  message(now(), ...)
 }
 
+mdebugf <- function(..., appendLF = TRUE,
+                    debug = getOption("future.debug", FALSE)) {
+  if (!debug) return()
+  message(now(), sprintf(...), appendLF = appendLF)
+}
+
+#' @importFrom utils capture.output
+mprint <- function(..., appendLF = TRUE, debug = getOption("future.debug", FALSE)) {
+  if (!debug) return()
+  message(paste(now(), capture.output(print(...)), sep = "", collapse = "\n"), appendLF = appendLF)
+}
+
+#' @importFrom utils capture.output str
+mstr <- function(..., appendLF = TRUE, debug = getOption("future.debug", FALSE)) {
+  if (!debug) return()
+  message(paste(now(), capture.output(str(...)), sep = "", collapse = "\n"), appendLF = appendLF)
+}
+
+
 ## From R.utils 2.0.2 (2015-05-23)
 hpaste <- function(..., sep="", collapse=", ", last_collapse=NULL,
                    max_head=if (missing(last_collapse)) 3 else Inf,
@@ -106,6 +117,9 @@ trim <- function(x, ...) {
   sub("[\t\n\f\r ]*$", "", sub("^[\t\n\f\r ]*", "", x))
 }
 
+comma <- function(x, sep = ", ") paste(x, collapse = sep)
+
+commaq <- function(x, sep = ", ") paste(sQuote(x), collapse = sep)
 
 import_from <- function(name, default = NULL, package) {
   ns <- getNamespace(package)
@@ -114,7 +128,7 @@ import_from <- function(name, default = NULL, package) {
   } else if (!is.null(default)) {
     default
   } else {
-    stop(sprintf("No such '%s' function: %s()", package, name))
+    stopf("No such '%s' function: %s()", package, name)
   }
 }
 
@@ -122,10 +136,6 @@ import_future <- function(name, default = NULL) {
   import_from(name, default = default, package = "future")
 }
 
-import_batchtools <- function(name, default = NULL) {
-  import_from(name, default = default, package = "batchtools")
-}
-
 ## Evaluates an expression in global environment.
 ## Because geval() is exported, we want to keep its environment()
 ## as small as possible, which is why we use local().  Without,
@@ -168,7 +178,7 @@ tempvar <- function(prefix = "var", value = NA, envir = parent.frame()) {
   }
 
   # Failed to find a unique temporary variable name
-  stop(sprintf("Failed to generate a unique non-existing temporary variable with prefix '%s'", prefix)) #nolint
+  stopf("Failed to generate a unique non-existing temporary variable with prefix '%s'", prefix) #nolint
 }
 
 
@@ -176,12 +186,38 @@ tempvar <- function(prefix = "var", value = NA, envir = parent.frame()) {
 result_has_errors <- function(result) {
   stop_if_not(inherits(result, "FutureResult"))
 
-  ## BACKWARD COMPATIBILITY: future (< 1.11.0)
-  if (inherits(result$condition, "error")) return(TRUE)
-  
   for (c in result$conditions) {
     if (inherits(c$condition, "error")) return(TRUE)
   }
   
   FALSE
 }
+
+
+
+#' @importFrom utils file_test
+file_info <- function(file) {
+  if (is.null(file) || is.na(file)) return("<NA>")
+  if (file_test("-f", file)) {
+    info <- sprintf("%d bytes", file.info(file)$size)
+    n <- length(readLines(file, warn = FALSE))
+    info <- sprintf("%s; %d lines", info, n)
+  } else {
+    info <- "<non-existing>"
+  }
+  sprintf("%s (%s)", sQuote(file), info)
+}
+
+
+assert_no_positional_args_but_first <- function(call = sys.call(sys.parent())) {
+  ast <- as.list(call)
+  if (length(ast) <= 2L) return()
+  ast <- ast[-(1:2)]
+  dots <- vapply(ast, FUN = identical, as.symbol("..."), FUN.VALUE = FALSE)
+  ast <- ast[!dots]
+  if (length(ast) == 0L) return()
+  names <- names(ast)
+  if (is.null(names) || any(names == "")) {    
+    stopf("Function %s() requires that all arguments beyond the first one are passed by name and not by position: %s", as.character(call[[1L]]), deparse(call, width.cutoff = 100L))
+  }
+}
